/*
 * Copyright (c) 1994 Hellmuth Michaelis
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Hellmuth Michaelis
 * 4. The name authors may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * @(#)ups.c, 1.20, Last Edit-Date: [Tue Aug  9 17:12:09 1994]
 *
 */

/*---------------------------------------------------------------------------*
 *
 *	Uninteruptable Power Supply driver
 *	----------------------------------
 *
 *	Hellmuth Michaelis, hm@hcshh.hcs.de
 *
 *	last edit-date: [Fri Aug  5 15:23:52 1994]
 *
 *	-hm	driver skeleton
 *	-hm	integration into FreeBSD
 *	-hm	bugfix port AC power bit
 *	-hm	more general interface
 *
 *---------------------------------------------------------------------------
 *
 *	This driver is for a Kyocera EP7000 machine's builtin
 *	uninterruptible power supply.
 *
 *	On locations 0xc06 and 0xc07 are registers belonging to a custom
 *	lsi circuit where one can get the machines status register from.
 *
 *	The documented bits are:
 *		bit 7 - battery status at power on time:
 * 			0 = battery not empty at power on time
 *			1 = battery empty at power on time
 *		bit 5 - battery empty status
 *			0 = battery is full
 *			1 = battery is empty
 *		bit 2 - current power supply status
 *			0 = working from battery power = powerfail
 *			1 = working from AC power
 *
 *	When the device is read with minor = 0, the caller gets the
 *	device independent status with the bits defined in upsbits.h.
 *
 *	When the device is read with minor = 1, the caller gets the
 *	untranslated status byte from the real hardware.
 *
 *---------------------------------------------------------------------------*/

#include "ups.h"
#if NUPS > 0

#include "param.h"
#include "buf.h"
#include "systm.h"
#include "ioctl.h"
#include "tty.h"
#include "proc.h"
#include "user.h"
#include "uio.h"
#include "kernel.h"
#include "malloc.h"
#include "i386/isa/isa_device.h"

#include "upsbits.h"

static int get_ups_stat(int port);

int upsprobe(), upsattach();
struct   isa_driver upsdriver = {upsprobe, upsattach, "ups"};

/* the port address */

static int port = 0;

/* ups status bit definitions in machine status byte */

#define KYO_POBEMPT	0x80	/* 1 = battery was empty on last pwr on */
#define KYO_BATEMPT	0x20	/* 1 = battery empty 			*/
#define KYO_ACPOWER	0x04	/* 1 = running from wall power		*/

#define UPSIOPORT	0xC06	/* c06 and c07 are the io ports for ups stat */

/*---------------------------------------------------------------------------*
 *	probe for hardware
 *---------------------------------------------------------------------------*/
int upsprobe(dev)
struct isa_device *dev;
{
	/* no way to probe it ! */
	
	return(1);
}

/*---------------------------------------------------------------------------*
 *	attach device
 *---------------------------------------------------------------------------*/
int upsattach(dev)
struct isa_device *dev;
{
	int ret;
	
	port = dev->id_iobase;

	if(port != UPSIOPORT)	/* MUST be this address !! */
	{
		port = 0;
		printf("ups%d: wrong port address!\n", dev->id_unit);		
		return (0);
	}
	ret = get_ups_stat(port);
	
	printf("ups%d: Kyocera EP7000 UPS driver 1.20, status byte = 0x%02x\n", dev->id_unit, ret);

	if(ret & KYO_POBEMPT)
		printf("ups%d: - battery was empty at power-up time\n", dev->id_unit);
	else
		printf("ups%d: - battery was full at power-up time\n", dev->id_unit);

	if(ret & KYO_BATEMPT)
		printf("ups%d: - battery is currently empty\n", dev->id_unit);
	else
		printf("ups%d: - battery is currently full\n", dev->id_unit);

	if(ret & KYO_ACPOWER)
		printf("ups%d: - system running from AC power\n", dev->id_unit);
	else
		printf("ups%d: - running from battery power, IN POWERFAIL !\n", dev->id_unit);
		
	return (1);
}

/*---------------------------------------------------------------------------*
 *	open device
 *---------------------------------------------------------------------------*/
int upsopen(int dev, int flag, int mode, struct proc *p)
{
	if(port != UPSIOPORT)	/* MUST be this address !! */
		return (ENXIO);	

	if((minor(dev) < 0) || (minor(dev) > 1))
		return (ENXIO);

	return 0;
}

/*---------------------------------------------------------------------------*
 *	close device
 *---------------------------------------------------------------------------*/
int upsclose(int dev, int flag, int mode, struct proc *p)
{
	return(0);
}

/*---------------------------------------------------------------------------*
 *	read from device
 *---------------------------------------------------------------------------*/
int upsread(dev, uio)
dev_t dev;
struct uio *uio;
{
	unsigned char upsbyte, ret;
	int error;
	
	if(port != UPSIOPORT)	/* MUST be this address !! */
		return (ENXIO);	

	upsbyte = get_ups_stat(port);

	if((minor(dev)) == 1)	 /* real hardware bits */
	{
		ret = upsbyte;	
	}
	else if((minor(dev)) == 0) /* translate hw stat -> general stat */
	{
		/* power source */
		
		if(upsbyte & KYO_ACPOWER)
			ret = 0;
		else
			ret = UPS_POWERFAIL;

		/* battery status at power on */
		
		if(upsbyte & KYO_POBEMPT)
			ret |= UPS_POBATEMPT;

		/* current battery status */

		if(upsbyte & KYO_BATEMPT)
			ret |= UPS_BATTEMPTY;

	}
	else
	{
		return(ENXIO);
	}
	
	if(error = uiomove(&ret, 1, uio))
		return (error);

	return(0);
}

/*---------------------------------------------------------------------------*
 *	get the UPS status byte
 *---------------------------------------------------------------------------*/
static int get_ups_stat(port)
int port;
{
	int status;
	
	outb(port,   0x48);	/* set protection index */
	outb(port+1, 0x47);	/* release protection */

	outb(port,   0x03);	/* set machine status index */
	status = inb(port+1);	/* get machine status */

	outb(port,   0x48);	/* set protection index */
	outb(port+1, 0x00);	/* set protection on */

	return(status);
}

#endif /* NUPS > 0 */

