/*
 * Copyright (c) 1994, 1995 Hellmuth Michaelis
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * @(#)upsd.c, 1.01, Last Edit-Date: [Tue Jul 18 21:58:22 1995]
 *
 */

/*---------------------------------------------------------------------------*
 *
 *	Uninteruptable Power Supply daemon
 *	----------------------------------
 *
 *	Hellmuth Michaelis, hm@hcshh.hcs.de
 *
 *	last edit-date: [Fri Aug  5 15:23:52 1994]
 *
 *	-hm	daemon start
 *	-hm	syslog messages
 *	-hm	protecting against multiple invocations
 *	-hm	sending message via wall to all users
 *	-hm	shutting the system down
 *	-hm	not reached code
 *	-hm	minor adjustments for FreeBSD 2.0.5
 *
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <varargs.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "upsd.h"
#include "upsbits.h"

u_char	log_status(int fd);
void	exit_func(void);
void	sig_func(int sig);
void	sig_child(int sig);
void	usage(void);
void	message(char *path);
void	attach(char *dev);
void	shutdown(void);

extern char *optarg;			/* option processing */
extern int optind;
extern int opterr;

char *pwfail = POWERFAIL_FILE;		/* powerfail text file */
char *pwgood = POWERGOOD_FILE;		/* power restored text file */

int powerfailstate = 0;			/* flag, in powerfail status */

/*---------------------------------------------------------------------------*
 *	main
 *---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	FILE	*fp;
	int	filed;
	int	childpid;
	int	mypid;
	int	nfiles;
	u_char	ups_stat;
	int	c;

	int	pf_seconds = 0;		/* how long in powerfail ? */
	int	pf_maxtime = PF_MAXTIME;/* maximum time running from DC */
	char 	*port = UPSDEVICE;	/* default value */
	int	time = CHECK_INT;	/* default value */
	
	/* parse the command line */

	while ((c = getopt(argc, argv, "d:t:f:g:m:")) != EOF)
	{
		switch (c)
		{
			case 'd':		/* device to read from */
				port = optarg;
				break;
				
			case 't':		/* delay between status reads */
				time = atoi(optarg);
				break;

			case 'm':		/* max seconds in powerfail */
				pf_maxtime = atoi(optarg);
				break;

			case 'f':		/* file containing fail text */
				pwfail = optarg;
				break;
				
			case 'g':		/* file containing ok text */
				pwgood = optarg;
				break;

			default:
				usage();	/* does not return */
				break;			
		}
	}

	mypid = getpid();
	
	/* check if another lock-file already exists */

	if((fp = fopen(UPSDPIDLOCK, "r")) != NULL)
	{
		/* lockfile found, check */
		
		int oldpid;

		/* read pid from file */
		
		if((fscanf(fp, "%d", &oldpid)) != 1)
		{
			fprintf(stderr, "upsd: Error, read pid from lockfile, terminating\n");
			exit(1);
		}

		/* check if process got from file is still alive */
		
		if((kill( oldpid, 0)) != 0)
		{
			/* process does not exist */

			/* close file */

			fclose(fp);

			/* remove file */

			fprintf(stderr, "upsd: removing old lock-file %s\n", UPSDPIDLOCK);

			unlink(UPSDPIDLOCK);
		}
		else
		{
			/* process is still alive */
			
			fprintf(stderr, "upsd: Error, another upsd is already running!\n");
			exit(1);
		}
	}
	
	/* ignore the terminal stop signals */

	signal(SIGTSTP, SIG_IGN);
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);

	/* if we were not started in the background,	*/
	/* fork and let the parent exit.		*/

	if((childpid = fork()) < 0)
	{
		fprintf(stderr, "upsd: Error, can't fork child, terminating\n");
		exit(1);
	}
	else if(childpid > 0)
		exit(0);	/* parent */

	/* write my pid into lock-file */
	
	if((fp = fopen(UPSDPIDLOCK, "w")) == NULL)
	{
		fprintf(stderr, "upsd: Error, can't open lockfile for wr, terminating\n");
		exit(1);
	}

	if((fprintf(fp, "%d", (int)getpid())) == EOF)
	{
		fprintf(stderr, "upsd: Error, can't write pid to lockfile, terminating\n");
		exit(1);
	}
	fclose(fp);

	/* disassociate from process group */

	if(setpgrp(0, getpid()) == -1)
	{
		fprintf(stderr, "upsd: Error, can't change process group, terminating\n");
		exit(1);
	}

	/* loose controlling tty */
	
	if((filed = open("/dev/tty", O_RDWR)) >= 0)
	{
		ioctl(filed, TIOCNOTTY, (char *)NULL);
		close(filed);
	}

	/* close all open files descriptors */

	nfiles = getdtablesize();
	
	for(filed = 0; filed < nfiles; filed++)
		close(filed);

	errno = 0;

	/* make sure we are not on a mounted filesystem */

	chdir("/");

	/* Clear any inherited file mode creation mask */

	umask(0);

	/* open path to syslogd */
	
	openlog(SYSL_IDENT, SYSL_OPTION, SYSL_FACILITY);

	/* register function to print final syslog message */
	
	atexit(exit_func);

	/* register all signal functions */
	
	signal(SIGHUP, sig_func);
	signal(SIGINT, sig_func);
	signal(SIGQUIT, sig_func);
	signal(SIGILL, sig_func);
	signal(SIGTRAP, sig_func);
	signal(SIGABRT, sig_func);
	signal(SIGIOT, sig_func);
	signal(SIGEMT, sig_func);
	signal(SIGFPE, sig_func);
	signal(SIGKILL, sig_func);	
	signal(SIGBUS, sig_func);
	signal(SIGSEGV, sig_func);
	signal(SIGSYS, sig_func);
	signal(SIGPIPE, sig_func);
	signal(SIGALRM, sig_func);
	signal(SIGTERM, sig_func);
	signal(SIGURG, sig_func);
	signal(SIGSTOP, sig_func);
	
	signal(SIGTSTP, SIG_IGN);	/* ignore */

	signal(SIGCONT, sig_func);

	signal(SIGCHLD, sig_child);	/* SIGCHLD handler */

	signal(SIGTTIN, SIG_IGN);	/* ignore */
	signal(SIGTTOU, SIG_IGN);	/* ignore */
	
	signal(SIGIO, sig_func);
	signal(SIGXCPU, sig_func);
	signal(SIGXFSZ, sig_func);
	signal(SIGVTALRM, sig_func);
	signal(SIGPROF, sig_func);
	signal(SIGWINCH, sig_func);
	signal(SIGINFO, sig_func);
	signal(SIGUSR1, sig_func);
	signal(SIGUSR2, sig_func);

	/* open the device to get status info from */
	
	if((filed = open(UPSDEVICE, O_RDONLY)) == -1)
	{
		syslog(LOG_ERR, "upsd: cannot open %s: %m, terminating", UPSDEVICE);
		exit(1);
	}	

	/* loop until forever */
	
	syslog(LOG_INFO, "upsd: device=%s, delay=%d, max=%d", port, time, pf_maxtime);
	
	for(;;)
	{
		ups_stat = log_status(filed);	/* get status from ups	*/

		sleep(time);			/* wait some time ..	*/

		if(powerfailstate)		/* in powerfail state ? */
		{
			pf_seconds += time;	/* how long in pwfail	*/

			if(pf_seconds >= pf_maxtime)	/* MAX ???	*/
			{
				close(filed);	/* close device		*/
				shutdown();	/* shut system down	*/
				exit(0);	/* terminate		*/
			}
		}
		else
		{
			pf_seconds = 0;		/* reset pwfail time */
		}
	}
}

/*---------------------------------------------------------------------------*
 *	return the ups status and handle syslog messages
 *---------------------------------------------------------------------------*/
unsigned char log_status(int fd)
{
	static int firsttime = 1;
	static int pobatstat, batstat;
	
	unsigned char upsbyte;
	
	/* read the current status byte from ups driver */

	if((read(fd, &upsbyte, 1)) != 1 )
	{
		syslog(LOG_ERR, "upsd: read failed from %s: %m, terminating", UPSDEVICE);
		exit(1);
	}

	if(firsttime)		/* first time */
	{
		if(upsbyte & UPS_POWERFAIL)
			syslog(LOG_INFO, "upsd: startup stat: POWERFAIL, using battery power!");
		else
			syslog(LOG_INFO, "upsd: startup stat: OK, using wall power");

		if(upsbyte & UPS_POBATEMPT)
		{
			syslog(LOG_INFO, "upsd: startup stat: Warning: Battery was low at power-up!");
			pobatstat = 1;
		}
		else
		{
			syslog(LOG_INFO, "upsd: startup stat: Battery OK at power-up");
			pobatstat = 0;
		}
		
		if(upsbyte & UPS_BATTEMPTY)
		{
			syslog(LOG_INFO, "upsd: startup stat: Warning: Battery is empty!");
			batstat = 1;
		}
		else
		{
			syslog(LOG_INFO, "upsd: startup stat: Battery is full");
			batstat = 0;
		}
			
		firsttime = 0;
	}
	else
	{
		/* check for running from battery, powerfail */		
#ifdef DEBUG
		syslog(LOG_INFO, "upsd: reading ups status [0x%x]", upsbyte);
#endif
		/* if condition changed since last wakeup, log it ! */

		if((upsbyte & UPS_POWERFAIL) && (powerfailstate == 0))
		{
			syslog(LOG_EMERG, "upsd: **** !!! POWERFAIL !!! *****");
			message(pwfail);
			powerfailstate = 1;
		}
		else if((!(upsbyte & UPS_POWERFAIL)) && (powerfailstate == 1))
		{
			syslog(LOG_EMERG, "upsd: End of powerfail condition!");
			message(pwgood);
			powerfailstate = 0;
		}

		if((upsbyte & UPS_POBATEMPT) && (pobatstat == 0))
		{
			syslog(LOG_INFO, "upsd: power on battery status change: Full -> Empty!");
			pobatstat = 1;
		}
		else if((!(upsbyte & UPS_POBATEMPT)) && (pobatstat == 1))
		{
			syslog(LOG_INFO, "upsd: power on battery status change: Empty -> Full!");
			pobatstat = 0;
		}

		if((upsbyte & UPS_BATTEMPTY) && (batstat == 0))
		{
			syslog(LOG_INFO, "upsd: battery status change: Full -> Empty!");
			batstat = 1;
		}
		else if((!(upsbyte & UPS_BATTEMPTY)) && (batstat == 1))
		{
			syslog(LOG_INFO, "upsd: battery status change: Empty -> Full!");
			batstat = 0;
		}
	}
	return(upsbyte);
}

/*---------------------------------------------------------------------------*
 *	registered function called at program exit
 *---------------------------------------------------------------------------*/
void exit_func(void)
{
	syslog(LOG_ERR, "upsd: terminating ...");
	unlink(UPSDPIDLOCK);
	closelog();
}

/*---------------------------------------------------------------------------*
 *	signal handler
 *---------------------------------------------------------------------------*/
void sig_func(int sig)
{
	syslog(LOG_ERR, "upsd: received signal %d", sig);
	exit(1);
}

/*---------------------------------------------------------------------------*
 *	sig_child signal handler
 *---------------------------------------------------------------------------*/
void sig_child(int sig)
{
	int pid;
	int status;

	while((pid = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0)
		;
}

/*---------------------------------------------------------------------------*
 *	usage
 *---------------------------------------------------------------------------*/
void usage(void)
{
	fprintf(stderr, "\nupsd - UPS monitor daemon Version 1.01\n");
	fprintf(stderr, "usage: upsd [-d device] [-t time] [-f path] [-g path]\n");
	fprintf(stderr, "       -d = ups monitor device         (default %s)\n", UPSDEVICE);
	fprintf(stderr, "       -t = delay between status reads (default %d seconds)\n", CHECK_INT);
	fprintf(stderr, "       -f = pathname for fail textfile (default %s)\n", POWERFAIL_FILE);
	fprintf(stderr, "       -g = pathname for good textfile (default %s)\n", POWERGOOD_FILE);
	fprintf(stderr, "       -m = max powerfail seconds      (default %d seconds)\n", PF_MAXTIME);
	exit(1);
}	

/*---------------------------------------------------------------------------*
 *	send message to all users
 *---------------------------------------------------------------------------*/
void message(char *path)
{
	struct stat st;
	
	/* message file must exist */

	if(stat(path, &st) == -1)
		return;

	/* and must be a regular file */
	
	if((st.st_mode & S_IFMT) != S_IFREG)
		return;

	if (!fork())
	{
		/* attach stdin/stdout/stderr to console */

		attach(CONSOLE);

		/* write message file to users */

		execlp(WALL, WALL, path, NULL);
	}
}

/*---------------------------------------------------------------------------*
 *	attach stdin/stdout/stderr to a device
 *---------------------------------------------------------------------------*/
void attach(char *dev)
{
	int     fd;

	close(0);		/* close stdin */
	close(1);		/* close stdout */
	close(2);		/* close stderr */

	fd = open(dev, O_RDWR);	/* open stdin */
	dup(fd);		/* open stdout */
	dup(fd);		/* open stderr */
}

/*---------------------------------------------------------------------------*
 *	shut the system down
 *---------------------------------------------------------------------------*/
void shutdown(void)
{
	int stat;
	
	syslog(LOG_EMERG, "upsd: Powerfail shutdown!");
	
	if (!fork()) {

		attach(CONSOLE);
		
		/* create a file "powerfail" in the root dir */

		if((stat = open(RCFILE, O_WRONLY | O_CREAT, 0644)) != -1)
			close(stat);

		/* execute shutdown command */

		if(stat != -1)
			execlp(SHELL, SHELL, "-c", "shutdown -r now", NULL);
		else
			execlp(SHELL, SHELL, "-c", "shutdown -h now", NULL);
	}
}

/* EOF */
